# scrapyd_manager/scrapyd_api.py
import os
import requests
from typing import List
from logging import getLogger
from django.db.models import Q
from .cache import ttl_cache
from . import models
from datetime import datetime

logger = getLogger(__name__)


def _auth_for_node(node: models.Node):
    """返回 node 的认证信息"""
    if getattr(node, "auth", False):
        return node.username, node.password
    return None


def start_spider(spider: models.Spider) -> bool:
    """启动爬虫并返回 Job"""
    url = f"{spider.version.project.node.url}/schedule.json"
    data = {
        "project": spider.version.project.name,
        "spider": spider.name,
    }
    resp = requests.post(url, data=data, auth=_auth_for_node(spider.version.project.node), timeout=15)
    resp.raise_for_status()
    result = resp.json()
    job_id = result.get("jobid")
    if not job_id:
        raise ValueError(f"爬虫启动失败：{result}")
    return True

def start_spiders(spiders: List[models.Spider]) -> bool:
    """批量启动爬虫"""
    for spider in spiders:
        start_spider(spider)
    return True


def stop_spider(spider: models.Spider) -> List[models.Job] | None:
    """停止某个爬虫的所有任务"""
    jobs = sync_jobs(spider.version.project.node)
    target_jobs = [j for j in jobs if j.spider == spider]
    stopped = []
    for job in target_jobs:
        stopped_job = stop_job(job)
        if stopped_job:
            stopped.append(stopped_job)
    return stopped or None


def stop_spiders(spiders: List[models.Spider]) -> List[models.Job] | None:
    """批量停止爬虫"""
    jobs = []
    for spider in spiders:
        job = stop_spider(spider)
        if job:
            jobs.extend(job)
    return jobs or None


def stop_job(job: models.Job) -> models.Job | None:
    """停止单个任务"""
    url = f"{job.spider.version.project.node.url}/cancel.json"
    data = {
        "project": job.spider.version.project.name,
        "job": job.job_id,
    }
    resp = requests.post(url, data=data, auth=_auth_for_node(job.spider.version.project.node), timeout=15)
    resp.raise_for_status()
    result = resp.json()
    if result.get("status") == "ok":
        job.status = "stopped"
        job.end_time = models.datetime.now()
        job.save(update_fields=["status", "end_time", "update_time"])
        return job
    return None


def stop_jobs(jobs: List[models.Job]) -> List[models.Job]:
    """批量停止任务"""
    stopped_jobs = []
    for job in jobs:
        stopped_job = stop_job(job)
        if stopped_job:
            stopped_jobs.append(stopped_job)
    return stopped_jobs


def start_spider_group(group: models.SpiderGroup) -> List[models.Job]:
    """启动任务组里的所有爬虫"""
    jobs = []
    for spider in group.spiders.all():
        job = start_spider(spider)
        if job:
            jobs.append(job)
    return jobs


def stop_spider_group(group: models.SpiderGroup) -> List[models.Job]:
    """停止任务组里的所有爬虫"""
    jobs = []
    for spider in group.spiders.all():
        job = stop_spider(spider)
        if job:
            jobs.extend(job)
    return jobs


@ttl_cache()
def get_job_info(job: models.Job) -> dict:
    """获取某个任务的详细信息"""
    url = f"{job.spider.version.project.node.url}/logs/{job.spider.version.project.name}/{job.spider.name}/{job.job_id}.json"
    resp = requests.get(url, auth=_auth_for_node(job.spider.version.project.node), timeout=15)
    resp.raise_for_status()
    return resp.json()


@ttl_cache()
def sync_jobs(node: models.Node) -> List[models.Job]:
    """列出节点上的所有任务并同步到数据库"""
    sync_node_projects(node)
    for project in models.Project.objects.filter(node=node):
        for version in project.versions.all():
            sync_project_version_spiders(version)
    url = f"{node.url}/listjobs.json"
    project_names = models.Project.objects.filter(node=node).values_list("name", flat=True).distinct()

    def match_job_spider(job_start_time: str, job_project_name, job_spider_name) -> models.Spider | None:
        valid_versions = []
        for project_version in models.ProjectVersion.objects.filter(project__node=node, project__name=job_project_name):
            if not project_version.version.isdigit():
                logger.warning("project version is not digit: %s@%s", project_version.version, job_project_name)
                continue
            version_datetime = datetime.fromtimestamp(int(project_version.version))
            job_datetime = datetime.strptime(job_start_time, "%Y-%m-%d %H:%M:%S.%f")
            if job_datetime > version_datetime:
                valid_versions.append(project_version)
        if not valid_versions:
            raise ValueError(f"no valid version found for job {job_project_name}@{job_spider_name} started at {job_start_time}")
        valid_versions.sort(key=lambda x: int(x.version), reverse=True)
        # 匹配spider
        for v in valid_versions:
            try:
                spider = models.Spider.objects.get(version=v, name=job_spider_name)
                break
            except models.Spider.DoesNotExist:
                continue
        else:
            raise ValueError(f"no valid project found for job {job_project_name}@{job_start_time}")
        return spider

    jobs = []
    for project_name in project_names:
        resp = requests.get(url, params={"project": project_name}, auth=_auth_for_node(node), timeout=15)
        resp.raise_for_status()
        data = resp.json()
        for status, entries in data.items():
            if status not in ("pending", "running", "finished"):
                continue

            for entry in entries:
                spider_name = entry["spider"]
                start_time = entry["start_time"]
                matched_spider = match_job_spider(start_time, project_name, spider_name)
                job = models.Job(
                    spider=matched_spider,
                    start_time=start_time,
                    job_id=entry["id"],
                    end_time=entry.get("end_time"),
                    items_url=entry.get("items_url"),
                    log_url=entry.get("log_url"),
                    pid=entry.get("pid"),
                    status=status,
                )
                job.gen_md5()
                jobs.append(job)

    deleted,  _ = models.Job.objects.filter(spider__version__project__node_id=node.id).delete()
    models.Job.objects.bulk_create(jobs)
    logger.info(f"deleted {deleted} old spiders, created {len(jobs)} new jobs for node {node}")
    return jobs


def sync_project_versions(project: models.Project):
    url = f"{project.node.url}/listversions.json"
    resp = requests.get(url, params={"project": project.name}, auth=_auth_for_node(project.node), timeout=15)
    resp.raise_for_status()
    data = resp.json()
    versions = data.get("versions", [])
    project_versions = []
    for version in versions:
        project_versions.append(models.ProjectVersion(project=project, version=version))
    models.ProjectVersion.objects.filter(~Q(version__in=versions), project=project).update(is_deleted=True)
    models.ProjectVersion.objects.bulk_create(project_versions, ignore_conflicts=True)
    logger.info(f"sync {len(project_versions)} versions for {project}")


def sync_node_projects(node: models.Node, include_version=True):
    """列出某个节点上的项目，支持是否展开版本"""
    url = f"{node.url}/listprojects.json"
    resp = requests.get(url, auth=_auth_for_node(node), timeout=15)
    resp.raise_for_status()
    data = resp.json()
    projects = data.get("projects", [])

    for project_name in projects:
        project, created = models.Project.objects.get_or_create(node=node, name=project_name)
        if include_version:
            sync_project_versions(project)
        # todo: 要删除已经移除的project
    logger.info(f"sync {len(projects)} project for {node}")


def sync_project_version_spiders(version: models.ProjectVersion) -> bool:
    """列出某个项目的爬虫"""
    if not version.is_spider_synced:
        url = f"{version.project.node.url}/listspiders.json"
        resp = requests.get(url, params={
            "project": version.project.name,
            "_version": version.version,
        }, auth=_auth_for_node(version.project.node), timeout=15)
        resp.raise_for_status()
        data = resp.json()
        spiders = data.get("spiders", [])
        results = [models.Spider(version=version, name=spider) for spider in spiders]
        models.Spider.objects.bulk_create(results)
        # 只需同步一次, 因为一个版本的spiders是不会变的
        logger.info(f"synced {len(spiders)} spiders for {version}@{version.project}")
        version.is_spider_synced = True
        version.save()
    return version.is_spider_synced


def add_version(version: models.ProjectVersion, egg_path: str):
    """部署新版本"""
    url = f"{version.project.node.url}/addversion.json"
    if not os.path.exists(egg_path):
        return {"status": "error", "reason": "egg file not found"}
    files = {"egg": open(egg_path, "rb")}
    data = {"project": version.project.name, "version": version.version}
    try:
        resp = requests.post(url, data=data, files=files, auth=_auth_for_node(version.project.node), timeout=60)
        resp.raise_for_status()
        return resp.json()
    finally:
        try:
            files["egg"].close()
        except Exception:
            pass


def delete_version(version: models.ProjectVersion):
    """删除某个版本"""
    url = f"{version.project.node.url}/delversion.json"
    data = {"project": version.project.name, "version": version.version}
    resp = requests.post(url, data=data, auth=_auth_for_node(version.project.node), timeout=15)
    resp.raise_for_status()
    return resp.json()


def delete_project(project: models.Project):
    """删除整个项目"""
    url = f"{project.node.url}/delproject.json"
    data = {"project": project.name}
    resp = requests.post(url, data=data, auth=_auth_for_node(project.node), timeout=15)
    resp.raise_for_status()
    return resp.json()


def daemon_status(node: models.Node) -> dict:
    """获取节点的 daemon 状态"""
    url = f"{node.url}/daemonstatus.json"
    resp = requests.get(url, auth=_auth_for_node(node), timeout=15)
    resp.raise_for_status()
    return resp.json()



# scrapyd_manager/models.py
from __future__ import annotations
from django.db import models
from django.utils.timezone import datetime
from .utils import get_md5


class Node(models.Model):
    name = models.CharField(max_length=100, verbose_name="节点名称", unique=True)
    ip = models.GenericIPAddressField(verbose_name="IP地址")
    port = models.IntegerField(default=6800, blank=True, null=True)
    ssl = models.BooleanField(default=False, verbose_name="是否启用SSL")
    description = models.CharField(max_length=500, blank=True, null=True, verbose_name="描述")
    auth = models.BooleanField(default=False, verbose_name="是否需要认证")
    username = models.CharField(max_length=255, blank=True, null=True)
    password = models.CharField(max_length=255, blank=True, null=True)
    create_time = models.DateTimeField(default=datetime.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @classmethod
    def default_node(cls):
        for node in cls.objects.all().prefetch_related("projects"):
            if node.projects.count():
                return node
        return None

    @property
    def default_project(self):
        for project in self.projects.all():
            if project.versions.count() > 0:
                return project
        return None

    @classmethod
    def default_project_of_node(cls, node: str | int | Node):
        if isinstance(node, str):
            assert node.isdigit(), "node id非法 必须是整型"
        if not isinstance(node, Node):
            node = Node.objects.get(id=int(node))
        return node.default_project

    class Meta:
        ordering = ["-create_time"]
        db_table = "scrapyd_node"
        verbose_name = verbose_name_plural = "Scrapyd Node"

    def __str__(self):
        return self.name

    @property
    def url(self):
        host = self.ip or "localhost"
        port = self.port or 6800
        return f"{"https" if self.ssl else "http"}://{host}:{port}"


class Project(models.Model):
    node = models.ForeignKey(Node, on_delete=models.DO_NOTHING, verbose_name="节点", db_constraint=False, related_name="projects")
    name = models.CharField(max_length=255, verbose_name="项目名")
    is_deleted = models.BooleanField(default=False, verbose_name="是否已被删除")
    create_time = models.DateTimeField(default=datetime.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def latest_version(self):
        return self.versions.order_by("-version").first()

    class Meta:
        db_table = "scrapy_project"
        verbose_name = verbose_name_plural = "Scrapy Project"
        unique_together = (("node", "name"),)

    def __str__(self):
        return self.name


class ProjectVersion(models.Model):
    project = models.ForeignKey(Project, on_delete=models.DO_NOTHING, verbose_name="项目", db_constraint=False, related_name="versions")
    version = models.CharField(max_length=255, verbose_name='版本')
    is_spider_synced = models.BooleanField(default=False, verbose_name="是否已同步当前版本爬虫")
    is_deleted = models.BooleanField(default=False, verbose_name="是否已被删除")
    create_time = models.DateTimeField(default=datetime.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def pretty(self):
        if self.version.isdigit():
            v = datetime.fromtimestamp(int(self.version))
        else:
            v = self.version
        return v

    @property
    def full_path(self):
        return f"{self.project.node.name}/{self.project.name}/{self.version}({self.pretty})"

    @property
    def short_path(self):
        return f"{self.version}({self.pretty})"

    class Meta:
        db_table = "scrapy_project_version"
        verbose_name = verbose_name_plural = "Scrapy Project Version"
        unique_together = (("project", "version"),)
        ordering = ["-version"]

    def __str__(self):
        return self.version


class Spider(models.Model):
    version = models.ForeignKey(ProjectVersion, on_delete=models.DO_NOTHING, verbose_name="版本", db_constraint=False, related_name="spiders")
    name = models.CharField(max_length=255, verbose_name="爬虫名称")
    create_time = models.DateTimeField(default=datetime.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    def __str__(self):
        return self.name

    class Meta:
        db_table = "scrapy_spider"
        verbose_name = verbose_name_plural = "Scrapy Spider"
        unique_together = (("version", "name"),)
        ordering = ["-version", "name"]


class SpiderGroup(models.Model):
    name = models.CharField(max_length=255, verbose_name="任务组名称", unique=True)
    node = models.ForeignKey(Node, db_constraint=False, on_delete=models.DO_NOTHING, verbose_name="节点")
    project = models.ForeignKey(Project, db_constraint=False, verbose_name='项目', on_delete=models.CASCADE)
    version = models.ForeignKey(ProjectVersion, verbose_name='版本', null=True, blank=True, on_delete=models.CASCADE)
    spiders = models.JSONField(default=list, verbose_name="爬虫")
    kwargs = models.JSONField(default=dict, verbose_name="参数", null=True, blank=True)
    description = models.CharField(max_length=200, blank=True, null=True, verbose_name="任务组描述")
    create_time = models.DateTimeField(default=datetime.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    @property
    def resolved_version(self):
        return self.version if self.version else self.project.latest_version

    @property
    def resolved_spiders(self):
        names = []
        for spider in self.spiders:
            names.append(spider["name"])
        spiders = Spider.objects.filter(name__in=names, version=self.resolved_version, version__project=self.project)
        return spiders

    class Meta:
        db_table = "scrapy_spider_group"
        verbose_name = verbose_name_plural = "Scrapy Spider Group"

    def __str__(self):
        return self.name


class StatusChoice(models.TextChoices):
    running = "running", "运行中"
    finished = "finished", "已结束"
    pending = "pending", "启动中"


class Job(models.Model):
    spider = models.ForeignKey(Spider, on_delete=models.DO_NOTHING, verbose_name="爬虫", db_constraint=False, related_name="jobs")
    job_id = models.CharField(max_length=255, verbose_name="任务ID")
    job_md5 = models.CharField(max_length=32, verbose_name="md5(job)", unique=True)
    start_time = models.DateTimeField(verbose_name="开始时间")
    end_time = models.DateTimeField(null=True, blank=True, verbose_name="结束时间")
    log_url = models.CharField(max_length=255, null=True, blank=True)
    items_url = models.CharField(max_length=255, null=True, blank=True)
    status = models.CharField(max_length=20, verbose_name="状态", choices=StatusChoice.choices)
    pid = models.IntegerField(null=True, blank=True, verbose_name="进程ID")
    create_time = models.DateTimeField(default=datetime.now, verbose_name="创建时间")
    update_time = models.DateTimeField(auto_now=True, verbose_name="更新时间")

    def gen_md5(self):
        if not self.job_md5:
            start_time = self.start_time
            if isinstance(start_time, datetime):
                start_time = start_time.strftime("%Y-%m-%d %H:%M:%S.%f")
            fields = [self.spider.version.project.name, self.spider.name, self.job_id, start_time]
            self.job_md5 = get_md5('-'.join(fields))
        return self.job_md5

    def save(self, *args, **kwargs):
        self.gen_md5()
        super().save(*args, **kwargs)

    class Meta:
        db_table = "scrapy_job"
        verbose_name = verbose_name_plural = "Scrapy Job"

    def __str__(self):
        return f"{self.spider.name} - {self.job_id} ({self.status})"




import time
from functools import wraps


_global_cache = {}


def ttl_cache(ttl: int = 60):
    """
    公共 TTL 缓存装饰器
    - 所有被装饰的函数共享同一个 cache 池
    - cache key = (函数名, 参数)
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # 以函数名+参数为 key，避免冲突
            key = (func.__name__, args, frozenset(kwargs.items()))
            now = time.time()

            # 命中缓存且未过期
            if key in _global_cache:
                result, expire_at = _global_cache[key]
                if expire_at > now:
                    return result

            # 重新计算并写入缓存
            result = func(*args, **kwargs)
            _global_cache[key] = (result, now + ttl)
            return result

        return wrapper
    return decorator



from django import forms
from django.contrib.admin import widgets
from . import models


class SpiderGroupForm(forms.ModelForm):
    spiders_select = forms.ModelMultipleChoiceField(
        queryset=models.Spider.objects.none(),
        widget=widgets.FilteredSelectMultiple("爬虫", is_stacked=False),
        required=False,
    )

    class Meta:
        model = models.SpiderGroup
        fields = "__all__"

    def __init__(self, *args, instance=None, **kwargs):
        if instance is None:
            instance = models.SpiderGroup()
            instance.node = models.Node.default_node()
            instance.project = instance.node.default_project if instance.node else None
        super().__init__(*args, instance=instance, **kwargs)
        instance: models.SpiderGroup = self.instance

        self.fields["project"].queryset = instance.node.projects.all()
        self.fields["version"].queryset = instance.project.versions.all()
        version = instance.version
        if not version:
            version = instance.project.latest_version
        if version:
            self.fields["version"].empty_label = f"自动最新版本[{version.short_path}]"
            spiders_select: forms.Field = self.fields["spiders_select"]
            spiders_select.queryset = version.spiders.all()
            spiders_select.initial = instance.resolved_spiders
        else:
            self.fields["version"].empty_label = "自动最新版本[暂无可用版本]"
        self.fields["version"].label_from_instance = lambda obj: obj.short_path

    def clean(self):
        cleaned_data = super().clean()
        spiders = []
        spiders_select = self.cleaned_data.pop("spiders_select")
        for spider in spiders_select:
            spiders.append({"name": spider.name})
        cleaned_data["spiders"] = spiders
        return cleaned_data

    def save(self, commit=True):
        obj: models.SpiderGroup = super().save(commit=False)
        obj.spiders = self.cleaned_data.get("spiders")
        if commit:
            obj.save()
        return obj



# Generated by Django 5.2.5 on 2025-09-02 19:19

import datetime
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Node',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, unique=True, verbose_name='节点名称')),
                ('ip', models.GenericIPAddressField(verbose_name='IP地址')),
                ('port', models.IntegerField(blank=True, default=6800, null=True)),
                ('ssl', models.BooleanField(default=False, verbose_name='是否启用SSL')),
                ('description', models.CharField(blank=True, max_length=500, null=True, verbose_name='描述')),
                ('auth', models.BooleanField(default=False, verbose_name='是否需要认证')),
                ('username', models.CharField(blank=True, max_length=255, null=True)),
                ('password', models.CharField(blank=True, max_length=255, null=True)),
                ('create_time', models.DateTimeField(default=datetime.datetime.now, verbose_name='创建时间')),
                ('update_time', models.DateTimeField(auto_now=True, verbose_name='更新时间')),
            ],
            options={
                'verbose_name': 'Scrapyd Node',
                'verbose_name_plural': 'Scrapyd Node',
                'db_table': 'scrapyd_node',
                'ordering': ['-create_time'],
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='项目名')),
                ('is_deleted', models.BooleanField(default=False, verbose_name='是否已被删除')),
                ('create_time', models.DateTimeField(default=datetime.datetime.now, verbose_name='创建时间')),
                ('update_time', models.DateTimeField(auto_now=True, verbose_name='更新时间')),
                ('node', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='projects', to='django_scrapyd_manager.node', verbose_name='节点')),
            ],
            options={
                'verbose_name': 'Scrapy Project',
                'verbose_name_plural': 'Scrapy Project',
                'db_table': 'scrapy_project',
                'unique_together': {('node', 'name')},
            },
        ),
        migrations.CreateModel(
            name='ProjectVersion',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', models.CharField(max_length=255, verbose_name='版本')),
                ('is_spider_synced', models.BooleanField(default=False, verbose_name='是否已同步当前版本爬虫')),
                ('is_deleted', models.BooleanField(default=False, verbose_name='是否已被删除')),
                ('create_time', models.DateTimeField(default=datetime.datetime.now, verbose_name='创建时间')),
                ('update_time', models.DateTimeField(auto_now=True, verbose_name='更新时间')),
                ('project', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='versions', to='django_scrapyd_manager.project', verbose_name='项目')),
            ],
            options={
                'verbose_name': 'Scrapy Project Version',
                'verbose_name_plural': 'Scrapy Project Version',
                'db_table': 'scrapy_project_version',
                'ordering': ['-version'],
                'unique_together': {('project', 'version')},
            },
        ),
        migrations.CreateModel(
            name='Spider',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, verbose_name='爬虫名称')),
                ('create_time', models.DateTimeField(default=datetime.datetime.now, verbose_name='创建时间')),
                ('update_time', models.DateTimeField(auto_now=True, verbose_name='更新时间')),
                ('version', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='spiders', to='django_scrapyd_manager.projectversion', verbose_name='版本')),
            ],
            options={
                'verbose_name': 'Scrapy Spider',
                'verbose_name_plural': 'Scrapy Spider',
                'db_table': 'scrapy_spider',
                'ordering': ['-version', 'name'],
                'unique_together': {('version', 'name')},
            },
        ),
        migrations.CreateModel(
            name='Job',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('job_id', models.CharField(max_length=255, verbose_name='任务ID')),
                ('job_md5', models.CharField(max_length=32, unique=True, verbose_name='md5(job)')),
                ('start_time', models.DateTimeField(verbose_name='开始时间')),
                ('end_time', models.DateTimeField(blank=True, null=True, verbose_name='结束时间')),
                ('log_url', models.CharField(blank=True, max_length=255, null=True)),
                ('items_url', models.CharField(blank=True, max_length=255, null=True)),
                ('status', models.CharField(choices=[('running', '运行中'), ('finished', '已结束'), ('pending', '启动中')], max_length=20, verbose_name='状态')),
                ('pid', models.IntegerField(blank=True, null=True, verbose_name='进程ID')),
                ('create_time', models.DateTimeField(default=datetime.datetime.now, verbose_name='创建时间')),
                ('update_time', models.DateTimeField(auto_now=True, verbose_name='更新时间')),
                ('spider', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.DO_NOTHING, related_name='jobs', to='django_scrapyd_manager.spider', verbose_name='爬虫')),
            ],
            options={
                'verbose_name': 'Scrapy Job',
                'verbose_name_plural': 'Scrapy Job',
                'db_table': 'scrapy_job',
            },
        ),
        migrations.CreateModel(
            name='SpiderGroup',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True, verbose_name='任务组名称')),
                ('spiders', models.JSONField(default=list, verbose_name='爬虫')),
                ('kwargs', models.JSONField(blank=True, default=dict, null=True, verbose_name='参数')),
                ('description', models.CharField(blank=True, max_length=200, null=True, verbose_name='任务组描述')),
                ('create_time', models.DateTimeField(default=datetime.datetime.now, verbose_name='创建时间')),
                ('update_time', models.DateTimeField(auto_now=True, verbose_name='更新时间')),
                ('node', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.DO_NOTHING, to='django_scrapyd_manager.node', verbose_name='节点')),
                ('project', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, to='django_scrapyd_manager.project', verbose_name='项目')),
                ('version', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='django_scrapyd_manager.projectversion', verbose_name='版本')),
            ],
            options={
                'verbose_name': 'Scrapy Spider Group',
                'verbose_name_plural': 'Scrapy Spider Group',
                'db_table': 'scrapy_spider_group',
            },
        ),
    ]



# scrapyd_manager/admin.py
from django.contrib import admin, messages
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from datetime import datetime
from django.utils.html import format_html
from django.urls import path
from django.shortcuts import redirect
from django.conf import settings
from . import models
from . import scrapyd_api
from . import forms


admin_project_name = "django_scrapyd_manager"
admin_prefix = settings.FORCE_SCRIPT_NAME if settings.FORCE_SCRIPT_NAME else ''
admin_prefix = f"{admin_prefix}/admin"


class CustomFilter(admin.SimpleListFilter):

    def choices(self, changelist):
        add_facets = changelist.add_facets
        facet_counts = self.get_facet_queryset(changelist) if add_facets else None
        for i, (lookup, title) in enumerate(self.lookup_choices):
            if add_facets:
                if (count := facet_counts.get(f"{i}__c", -1)) != -1:
                    title = f"{title} ({count})"
                else:
                    title = f"{title} (-)"
            yield {
                "selected": self.value() == str(lookup),
                "query_string": changelist.get_query_string(
                    {self.parameter_name: lookup}
                ),
                "display": title,
            }

    def queryset(self, request, queryset):
        value = self.value()
        if value:
            return queryset.filter(**{self.parameter_name: value})
        return queryset


@admin.register(models.Node)
class NodeAdmin(admin.ModelAdmin):
    list_display = ("name", "linked_url", "description", "related_projects", "auth", "daemon_status", "create_time")
    readonly_fields = ("create_time", "update_time")

    def linked_url(self, obj: models.Node) -> str:
        return format_html(f"<a href='{obj.url}'>{obj.url}</a>")
    linked_url.short_description = "Scrapyd地址"

    def daemon_status(self, obj: models.Node):
        return scrapyd_api.daemon_status(obj).get("status") == "ok"
    daemon_status.short_description = "状态"
    daemon_status.boolean = True

    def related_projects(self, obj: models.Node):
        projects = []
        for project in obj.projects.all()[:5]:
            href = f"{admin_prefix}/{admin_project_name}/{models.Project._meta.model_name}/?node_id={obj.id}"
            projects.append(f"<a href='{href}'>{project.name}</a>")
        if len(projects) == 5:
            projects.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(projects))
    related_projects.short_description = "项目"


class ProjectNodeFilter(CustomFilter):
    """右侧过滤：Node（节点）"""
    title = "节点"
    parameter_name = "node_id"

    def lookups(self, request, model_admin):
        return [(str(n.id), n.name) for n in models.Node.objects.all().order_by("name")]


class ProjectFilter(CustomFilter):
    """右侧过滤：Project（项目）"""
    title = "项目"
    parameter_name = "name"

    node_filter = ProjectNodeFilter

    def lookups(self, request, model_admin):
        node_id = request.GET.get(self.node_filter.parameter_name)
        if node_id:
            projects = models.Project.objects.filter(node_id=node_id).values_list("name", flat=True).distinct().order_by("name")
            return [(name, name) for name in projects]
        return []


@admin.register(models.Project)
class ProjectAdmin(admin.ModelAdmin):
    list_display = ("name", "latest_version", "related_versions", "is_deleted", "create_time")
    readonly_fields = ("create_time", "update_time")
    list_filter = (ProjectNodeFilter, )

    def has_change_permission(self, request, obj = ...):
        return False

    def has_delete_permission(self, request, obj = ...):
        return False

    def latest_version(self, obj: models.Project):
        version = obj.versions.order_by("-version").first()
        if version:
            href = f"{admin_prefix}/{admin_project_name}/{models.ProjectVersion._meta.model_name}/?id={version.id}"
            return format_html(f'<a href="{href}">{version.pretty}</a>')
        return '-'
    latest_version.short_description = "最新版本"

    def related_versions(self, obj: models.Project):
        href = f"{admin_prefix}/{admin_project_name}/{models.ProjectVersion._meta.model_name}/?project_id={obj.id}"
        return format_html(f'<a href="{href}">{obj.versions.count()}</a>')
    related_versions.short_description = "版本数量"

    def get_object(self, request, object_id, from_field = ...):
        return get_object_or_404(models.Project, pk=object_id)

    def changelist_view(self, request, extra_context=None):

        node_id = request.GET.get(ProjectNodeFilter.parameter_name)
        if node_id:
            node = get_object_or_404(models.Node, pk=node_id)
        else:
            node = models.Node.objects.order_by('name').first()
        if node:
            params = request.GET.copy()
            params[ProjectNodeFilter.parameter_name] = str(node.id)
            request.GET = params
            scrapyd_api.sync_node_projects(node)
        return super().changelist_view(request, extra_context)


class VersionNodeFilter(ProjectNodeFilter):
    # 使用与 Django 内置 FieldListFilter 一致的参数名，便于复用已有的默认逻辑
    parameter_name = "project__node_id"


class VersionProjectFilter(ProjectFilter):
    """右侧过滤：Project（按项目名，不含版本），受 Node 选择联动"""
    parameter_name = "project__name"
    node_filter = VersionNodeFilter


@admin.register(models.ProjectVersion)
class ProjectVersionAdmin(admin.ModelAdmin):
    list_display = ("id", "linked_version", "project", "spider_count", "is_spider_synced", "is_deleted", "create_time")
    readonly_fields = ("create_time", "update_time")
    list_filter = (VersionNodeFilter, VersionProjectFilter)

    def has_change_permission(self, request, obj = ...):
        return False

    def has_delete_permission(self, request, obj = ...):
        return False

    def linked_version(self, obj: models.ProjectVersion):
        if obj.version.isdigit():
            version_datetime = datetime.fromtimestamp(int(obj.version))
        else:
            version_datetime = obj.version
        href = f"{admin_prefix}/{admin_project_name}/{models.Spider._meta.model_name}/?version__version={obj.version}"
        return format_html(f'<a href="{href}">{obj.version}({version_datetime})</a>')
    linked_version.admin_order_field = "version"
    linked_version.short_description = "版本"

    def spider_count(self, obj: models.ProjectVersion):
        return obj.spiders.count()
    spider_count.short_description = "爬虫数量"

    def changelist_view(self, request, extra_context=None):
        node_id = request.GET.get(VersionProjectFilter.node_filter.parameter_name)
        project_name = request.GET.get(VersionProjectFilter.parameter_name)
        if node_id:
            node = get_object_or_404(models.Node, pk=node_id)
        else:
            node = models.Node.objects.order_by('name').first()
        if node:
            scrapyd_api.sync_node_projects(node)
            if not project_name:
                project = models.Project.objects.filter(node=node).order_by("name").first()
                if project:
                    project_name = project.name
            q = request.GET.copy()
            q[VersionProjectFilter.node_filter.parameter_name] = str(node.id)
            q[VersionProjectFilter.parameter_name] = project_name
            request.GET = q
        return super().changelist_view(request, extra_context)


class SpiderNodeFilter(ProjectNodeFilter):
    # 使用与 Django 内置 FieldListFilter 一致的参数名，便于复用已有的默认逻辑
    parameter_name = "version__project__node_id"


class SpiderProjectFilter(ProjectFilter):
    """右侧过滤：Project（按项目名，不含版本），受 Node 选择联动"""
    parameter_name = "version__project__name"
    node_filter = SpiderNodeFilter


class SpiderProjectVersionFilter(CustomFilter):
    title = "版本"
    parameter_name = "version__version"

    def lookups(self, request, model_admin):
        node_id = request.GET.get(SpiderNodeFilter.parameter_name)
        project_name = request.GET.get(SpiderProjectFilter.parameter_name)
        if not (project_name and node_id):
            return []
        # 获取当前节点下的所有项目版本
        versions = models.ProjectVersion.objects.filter(
            project__node_id=node_id,
            project__name=project_name,
        ).values_list('version', flat=True).order_by("-version")
        return [(v, f'{v}({datetime.fromtimestamp(int(v))})') for v in versions]


@admin.register(models.Spider)
class SpiderAdmin(admin.ModelAdmin):
    list_display = ("name", "project_name", "project_node_name", "start_spider", "create_time")
    readonly_fields = ("create_time", "update_time")
    list_filter = (SpiderNodeFilter, SpiderProjectFilter, SpiderProjectVersionFilter)
    actions = ["start_spiders"]

    def has_change_permission(self, request, obj = ...):
        return False

    def has_delete_permission(self, request, obj = ...):
        return False

    def get_urls(self):

        urls = super().get_urls()
        custom_urls = [
            path(
                "<path:spider_id>/start/",
                self.admin_site.admin_view(self.start_spider_view),
                name="scrapy_spider_start",
            ),
        ]
        return custom_urls + urls

    def start_spider_view(self, request, spider_id):
        spider = get_object_or_404(models.Spider, pk=spider_id)
        try:
            job_id = scrapyd_api.start_spider(spider)
            self.message_user(request, f"成功启动爬虫 {spider.name} (job_id={job_id})", level=messages.SUCCESS)
        except Exception as e:
            self.message_user(request, f"启动失败: {e}", level=messages.ERROR)
        from django.shortcuts import redirect
        return redirect(request.META.get("HTTP_REFERER", f"{admin_prefix}/{admin_project_name}/spider/"))

    def project_name(self, obj: models.Spider):
        return obj.version.project.name
    project_name.admin_order_field = "name"
    project_name.short_description = "项目名称"

    def project_node_name(self, obj: models.Spider):
        return obj.version.project.node.name
    project_node_name.admin_order_field = "project__node__name"
    project_node_name.short_description = "节点名称"

    def get_queryset(self, request):
        node_id = request.GET.get(SpiderNodeFilter.parameter_name)
        project_name = request.GET.get(SpiderProjectFilter.parameter_name)
        version = request.GET.get(SpiderProjectVersionFilter.parameter_name)
        if node_id and project_name and version:
            project_version = models.ProjectVersion.objects.filter(
                project__node_id=node_id, project__name=project_name, version=version,
            ).first()
            if project_version:
                scrapyd_api.sync_project_version_spiders(project_version)
                return models.Spider.objects.filter(version=project_version)
            return models.Spider.objects.none()
        return super().get_queryset(request).prefetch_related("version", "version__project")

    def get_object(self, request, object_id, from_field = ...):
        return get_object_or_404(models.Spider, pk=object_id)

    def changelist_view(self, request, extra_context=None):
        if request.GET.get("id"):
            return super().changelist_view(request, extra_context)
        node_id = request.GET.get(SpiderNodeFilter.parameter_name)
        project_name = request.GET.get(SpiderProjectFilter.parameter_name)
        version = request.GET.get(SpiderProjectVersionFilter.parameter_name)
        last_node_id = request.COOKIES.get("last_node_id")
        last_project_name = request.COOKIES.get("last_project_name")
        new_query = request.GET.copy()
        if last_node_id and last_node_id != node_id:
            new_query.pop(SpiderProjectFilter.parameter_name, None)
            new_query.pop(SpiderProjectVersionFilter.parameter_name, None)
            version = project_name = None
        elif last_project_name and last_project_name != project_name:
            new_query.pop(SpiderProjectVersionFilter.parameter_name, None)
            version = None
        if not node_id:
            node_id = models.Node.objects.order_by('name').values_list("id", flat=True).first()
        if node_id:
            new_query[SpiderNodeFilter.parameter_name] = str(node_id)
            if not project_name:
                project = models.Node.default_project_of_node(node_id)
                if project:
                    project_name = project.name if project else None
                    new_query[SpiderProjectFilter.parameter_name] = project_name
                    if not version:
                        version = project.latest_version
                        if version:
                            new_query[SpiderProjectVersionFilter.parameter_name] = version.version
        request.GET = new_query
        response = super().changelist_view(request, extra_context)
        response.set_cookie("last_node_id", node_id)
        response.set_cookie("last_project_name", project_name)
        return response

    def start_spider(self, obj: models.Spider):
        return format_html(
            '<a class="button" href="{}">启动</a>',
            f"{admin_prefix}/{admin_project_name}/{models.Spider._meta.model_name}/{obj.id}/start/"
        )
    start_spider.short_description = "运行"

    def start_spiders(self, request, queryset):
        """启动选中的爬虫"""
        if not queryset:
            messages.error(request, "请选择要启动的爬虫")
            return
        for spider in queryset:
            try:
                job_id = scrapyd_api.start_spider(spider)
                messages.success(request, f"成功启动爬虫 {spider.name} (job_id={job_id})")
            except Exception as e:
                messages.error(request, f"启动爬虫 {spider.name} 失败: {str(e)}")

    start_spiders.short_description = "启动选中的爬虫"


@admin.register(models.SpiderGroup)
class SpiderGroupAdmin(admin.ModelAdmin):
    list_display = ("name", "node", "project", "related_spiders", "formatted_kwargs", "formatted_version", "start_spider_group", "create_time")
    readonly_fields = ("create_time", "update_time")
    # filter_horizontal = ("nodes",)
    form = forms.SpiderGroupForm
    actions = ["start_group_spiders"]
    list_filter = ("node",)
    fields = (
        "name",
        ("node", "project"),
        "version",
        "spiders_select",
        "description",
        "kwargs",
        "create_time",
        "update_time",
    )

    def get_queryset(self, request):
        return super().get_queryset(request).select_related("project", "node", "version", "version__project")

    def formatted_version(self, obj: models.SpiderGroup):
        if obj.version:
            return obj.version
        latest = obj.project.latest_version
        if latest:
            return f"自动最新版本[{latest.short_path}]"
        return "自动最新版本[暂无可用版本]"

    formatted_version.short_description = "版本"

    def formatted_kwargs(self, obj: models.SpiderGroup):
        args = []
        for key, value in obj.kwargs.items():
            args.append(f"{key} = {value}")
        if len(args) == 5:
            args.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(args))
    formatted_kwargs.short_description = "参数"

    def related_spiders(self, obj: models.SpiderGroup):
        spiders = []
        for spider in obj.resolved_spiders:
            href = f"{admin_prefix}/{admin_project_name}/{models.Spider._meta.model_name}/?id={spider.id}"
            spiders.append(f"<a href='{href}'>{spider.name}</a>")
        if len(spiders) == 5:
            spiders.append("···")
        return format_html('<span style="line-height: 1">%s</span>' % '<br>'.join(spiders))
    related_spiders.short_description = "爬虫"

    def start_spider_group(self, obj: models.SpiderGroup):
        return format_html(
            '<a class="button" href="{}">启动</a>',
            f"{admin_prefix}/{admin_project_name}/{models.SpiderGroup._meta.model_name}/{obj.id}/start/"
        )
    start_spider_group.short_description = "运行"

    def start_group_spiders(self, request, queryset):
        """启动选中的爬虫组（组内所有爬虫）"""
        if not queryset:
            messages.error(request, "请选择要启动的爬虫组")
            return
        for group in queryset:
            spiders = group.resolved_spiders
            if not spiders:
                messages.warning(request, f"爬虫组 {group.name} 内没有爬虫")
                continue
            for spider in spiders:
                try:
                    job_id = scrapyd_api.start_spider(spider)
                    messages.success(request, f"组 {group.name} -> 启动爬虫 {spider.name} (job_id={job_id})")
                except Exception as e:
                    messages.error(request, f"组 {group.name} -> 启动爬虫 {spider.name} 失败: {str(e)}")
                    break

    start_group_spiders.short_description = "启动选中的爬虫组"

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "<path:group_id>/start/",
                self.admin_site.admin_view(self.start_group_view),
                name="scrapy_spidergroup_start",
            ),
            path(
                "api/node/<int:node_id>/projects/",
                self.admin_site.admin_view(self.get_projects),
                name="node_projects",
            ),
            path(
                "api/project/<int:project_id>/versions/",
                self.admin_site.admin_view(self.get_versions),
                name="project_versions",
            ),
            path(
                "api/version/spiders/",
                self.admin_site.admin_view(self.get_spiders),
                name="version_spiders",
            ),
        ]
        return custom_urls + urls

    @staticmethod
    def get_projects(request, node_id):
        projects = models.Project.objects.filter(node_id=node_id)
        data = [{"id": p.id, "text": p.name} for p in projects]
        return JsonResponse(data, safe=False)

    @staticmethod
    def get_versions(request, project_id):
        if project_id:
            versions = models.ProjectVersion.objects.filter(project_id=project_id).order_by("-create_time")
        else:
            node_id = request.GET.get("node_id")
            versions = models.ProjectVersion.objects.filter(project__node_id=node_id).order_by("-create_time")
        data = [{"id": v.id, "text": f"{v.version} ({v.pretty})"} for v in versions]
        return JsonResponse(data, safe=False)

    @staticmethod
    def get_spiders(request):
        version_id = request.GET.get("version_id")
        if version_id:
            spiders = models.Spider.objects.filter(version_id=version_id)
        else:
            node_id = request.GET.get("node_id")
            project_id = request.GET.get("project_id")
            spiders = models.Spider.objects.filter(version__project_id=project_id, version__project__node_id=node_id)
        data = [{"id": s.id, "text": s.name} for s in spiders]
        return JsonResponse(data, safe=False)

    def start_group_view(self, request, group_id):
        group = get_object_or_404(models.SpiderGroup, pk=group_id)
        spiders = group.resolved_spiders
        if not spiders:
            self.message_user(request, f"组 {group.name} 内没有爬虫", level=messages.WARNING)
        for spider in spiders:
            try:
                job_id = scrapyd_api.start_spider(spider)
                self.message_user(request, f"组 {group.name} -> 启动爬虫 {spider} (job_id={job_id})",
                                  level=messages.SUCCESS)
            except Exception as e:
                self.message_user(request, f"组 {group.name} -> 启动爬虫 {spider.name} 失败: {e}",
                                  level=messages.ERROR)
        return redirect(request.META.get("HTTP_REFERER", f"{admin_prefix}/{admin_project_name}/{models.SpiderGroup._meta.model_name}/"))

    class Media:
        js = ("admin/js/core.js", "js/spider_group_linked.js")


class JobNodeFilter(ProjectNodeFilter):
    parameter_name = "spider__version__project__node_id"


class JobProjectFilter(ProjectFilter):
    parameter_name = "spider__version__project__name"
    node_filter = JobNodeFilter


class JobSpiderFilter(CustomFilter):
    """右侧过滤：Spider（爬虫）"""
    title = "爬虫"
    parameter_name = "spider_id"

    def lookups(self, request, model_admin):
        node_id = request.GET.get(JobNodeFilter.parameter_name)
        project_name = request.GET.get(JobProjectFilter.parameter_name)
        if node_id and project_name:
            spiders = models.Spider.objects.filter(project__node_id=node_id, project__name=project_name).order_by("name")
            return [(str(s.id), s.name) for s in spiders]
        return []


class JobStatusFilter(CustomFilter):
    parameter_name = "status"
    title = "状态"

    def lookups(self, request, model_admin):
        filters = [
            (x, models.StatusChoice[x].label) for x in models.Job.objects.values_list("status", flat=True).distinct()
        ]
        filters.sort()
        return filters


@admin.register(models.Job)
class JobAdmin(admin.ModelAdmin):
    list_display = (
        "job_id", "job_project_version", "job_spider", "start_time", "end_time", "status", "pid", "stop_job",
    )
    readonly_fields = ("create_time", "update_time", "start_time", "end_time", "pid", "log_url", "items_url", "spider", "status")
    list_filter = (JobStatusFilter, JobNodeFilter, JobProjectFilter)
    actions = ["start_jobs", "stop_jobs"]
    ordering = ("-status", "-start_time")

    def job_node(self, obj):
        return obj.spider.version.project.node.name
    job_node.admin_order_field = "spider__version__project__node__name"
    job_node.short_description = "节点名称"

    def job_project(self, obj):
        return obj.spider.version.project.name
    job_project.admin_order_field = "spider__version__project__name"
    job_project.short_description = "项目名称"

    def job_project_version(self, obj):
        return obj.spider.version.version
    job_project_version.admin_order_field = "spider__version__version"
    job_project_version.short_description = "项目版本"

    def job_spider(self, obj):
        return obj.spider.name
    job_spider.admin_order_field = "spider__name"
    job_spider.short_description = "爬虫名称"

    def start_jobs(self, request, queryset):
        """启动选中的 Job 对应的爬虫（重新运行）"""
        if not queryset:
            messages.error(request, "请选择要启动的任务")
            return
        for job in queryset:
            try:
                job_id = scrapyd_api.start_spider(job.spider)
                messages.success(request, f"成功重新启动任务 {job.spider.name} (job_id={job_id})")
            except Exception as e:
                messages.error(request, f"重新启动任务 {job.spider.name} 失败: {str(e)}")
    start_jobs.short_description = "重新启动选中的任务"

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "<path:job_id>/stop/",
                self.admin_site.admin_view(self.stop_job_view),
                name="scrapy_job_stop",
            ),
        ]
        return custom_urls + urls

    def stop_job_view(self, request, job_id):
        job = get_object_or_404(models.Job, pk=job_id)
        try:
            scrapyd_api.stop_job(job)
            self.message_user(request, f"成功停止任务 {job.job_id} ({job.spider.name})", level=messages.SUCCESS)
        except Exception as e:
            self.message_user(request, f"停止任务失败: {e}", level=messages.ERROR)
        return redirect(request.META.get("HTTP_REFERER", f"{admin_prefix}/{admin_project_name}/{models.Job._meta.model_name}/"))

    def stop_job(self, obj: models.Job):
        if obj.status == "running":  # 可选: 只在运行中显示按钮
            return format_html(
                '<a class="button" href="{}">停止</a>',
                f"{admin_prefix}/{admin_project_name}/{models.Job._meta.model_name}/{obj.id}/stop/"
            )
        return "-"
    stop_job.short_description = "操作"

    def stop_jobs(self, request, queryset):
        """停止选中的爬虫任务"""
        if not queryset:
            messages.error(request, "请选择要停止的任务")
            return
        for job in queryset:
            try:
                scrapyd_api.stop_job(job)
                messages.success(request, f"成功停止任务 {job.job_id} ({job.spider.name})")
            except Exception as e:
                messages.error(request, f"停止任务 {job.job_id} 失败: {str(e)}")
    stop_jobs.short_description = "停止选中的爬虫任务"

    def get_queryset(self, request):
        node_id = request.GET.get(JobNodeFilter.parameter_name)
        if node_id:
            node = get_object_or_404(models.Node, pk=node_id)
        else:
            return models.Job.objects.none()
        scrapyd_api.sync_jobs(node)
        return super().get_queryset(request).prefetch_related("spider", "spider__version", "spider__version__project")

    def get_object(self, request, object_id, from_field = ...):
        return get_object_or_404(models.Job, pk=object_id)

    def changelist_view(self, request, extra_context=None):
        if JobNodeFilter.parameter_name not in request.GET:
            default_node = models.Node.objects.order_by('name').first()
            if default_node:
                q = request.GET.copy()
                q[JobNodeFilter.parameter_name] = str(default_node.id)
                project_name = request.GET.get(JobProjectFilter.parameter_name)
                if project_name:
                    default_project = models.Project.objects.filter(node_id=default_node.id, name=project_name).first()
                else:
                    default_project = models.Project.objects.filter(node_id=default_node.id).first()
                if default_project:
                    q[JobProjectFilter.parameter_name] = default_project.name
                request.GET = q
        return super().changelist_view(request, extra_context)
